/* Process helpers for finding windows in Windows.
 Created by nugarp/spazzy. Last Updated 9/25/22. */

const memoryjs = require('memoryjs');
const _ = require('lodash');
const cp = require('child_process');
const md5File = require('md5-file');

let findGame = (pair, prox) => {
    let procs = prox;
    let exeFile = pair.app;
    let title = pair.title;
    return new Promise((resolve, reject) => {
        let results = _.filter(procs, p => p.szExeFile.toLowerCase() == exeFile.toLowerCase());
        let found = null;
        if (results == null || results.length === 0) { 
            resolve(null);
            return;
        } else {
            _.each(results, (p) => {
                if (found == null) {
                    let stdout = cp.execSync(`(tasklist /FI "PID eq ${p.th32ProcessID}" /fo list /v)`).toString();
                    let execDetails = {};
                    stdout.split("\r\n").forEach(line => {
                        let param = [line.substring(0, line.indexOf(":")), line.substring(line.indexOf(":") + 1)] // split doesnt work if there is ":" in the title for ex: ]line.split(":")
                        if (!param[1]) return
                        execDetails[param[0].trim()] = param[1].trim()
                    })

                    const details = {
                        ...p,
                        ImageName: execDetails["Image Name"],
                        PID: Number(execDetails["PID"]),
                        SessionName: execDetails["Session Name"],
                        Session: Number(execDetails["Session#"]),
                        MemUsage: execDetails["Mem Usage"],
                        Status: execDetails["Status"],
                        UserName: execDetails["User Name"],
                        CPUTime: execDetails["CPU Time"],
                        WindowTitle: execDetails["Window Title"]
                    }

                    // debugging output - uncomment if needed.
                    // console.log("CHECKING OUT " + JSON.stringify({
                    //     ImageName: execDetails["Image Name"],
                    //     PID: Number(execDetails["PID"]),
                    //     SessionName: execDetails["Session Name"],
                    //     Session: Number(execDetails["Session#"]),
                    //     MemUsage: execDetails["Mem Usage"],
                    //     Status: execDetails["Status"],
                    //     UserName: execDetails["User Name"],
                    //     CPUTime: execDetails["CPU Time"],
                    //     WindowTitle: execDetails["Window Title"]
                    // }))

                    if (!title || title.length == 0 || (details.WindowTitle || "").toLowerCase().indexOf(title.toLowerCase()) !== -1) {
                        found = details;
                    }
                }
            })
        }

        if (found !== null) {
            let pathinfotxt = cp.execSync(`wmic process where "ProcessID like ${found.PID}" get ExecutablePath`).toString(); //ProcessID, ExecutablePath /format:csv`).toString();
            let pathinfoarr = pathinfotxt.split("\r\n");
            let exePath = pathinfoarr.length > 1 ? pathinfoarr[1].trim() : null;

            let ret = {
                exe: found.szExeFile,
                exePath: exePath,
                title: found.WindowTitle,
                pid: found.PID,
                th32ProcessID: found.th32ProcessID,
                md5: null
            };
	    console.log(JSON.stringify(ret));

            try {
                ret.process = memoryjs.openProcess(ret.pid);
            } catch(e) {
                // couldn't access exe, maybe didn't have admin permissions?
                console.log("ERROR - could not attach to process. Did you run the game as admin but not this software?");
                resolve(null);
                return;
            }
            resolve({ target: ret, cfg: pair });
            return;
        }
        resolve(null);
        return;
    });
}

let findGames = (arr) => { // pairs of (app/exeFile, title/title)
    let procs = memoryjs.getProcesses();
    let promises = [];
    let found = null;
    _.each(arr, (pair) => {
        promises.push(findGame(pair, procs));
    });
    return Promise.all(promises).then((results) => {
        _.each(results, r => {
            if (found === null && r !== null) {
                found = r;
            }
        });
        return found;
    });
}

exports.findGames = findGames