/* Feedback plugin for Deadstorm, v0.1.2.
 Created by nugarp/spazzy. Last Updated 10/03/22.
 Big thanks to argonlefou for help with the
 memory addresses. No requirement on loader but
 only tested with TP.

 v0.1.2 - Force set autofire timings in hook,
  and made the grenade affect both players when
  it explodes. Also fixed typo for p2 grenade.
 v0.1.1 - Fixed typo with "died" condition.
 */

const { Plugin } = require('./plugin');
const { TYPE } = require('../core/memoryHelpers');

class DeadstormPlugin extends Plugin {

    // these variables are updated in hook
    #outputsAddress = 0x301348658;
    // #playerStructPtrAddress  = 0x0;
    // #outputsPtrAddress = 0x0;
    // #gameModePtrAddress = 0x0;

    // temporary lastState to use at initialization
    #lastState = {
        // p1Vibration: -1,
        // p2Vibration: -1
        // mode1: -1,
        // life1: -1,
        // ammo1: -1,
        // gren1: -1,
        // mode2: -1,
        // life2: -1,
        // ammo2: -1,
        // gren2: -1,
        // motor1: 0,
        // motor2: 0
    };

    #p1FireStart = null;
    #p2FireStart = null;
    #p1LastSwitch = null;
    #p2LastSwitch = null;
    #p1Weapon = 0; // 0 = gun, 1 = rapid, 2 = cannon
    #p2Weapon = 0;

    constructor(proc, players, mem, cfg) {
        console.log('Deadstorm (BETA): loaded');
        super(proc, players, mem, cfg);
    }

    hook () {
        console.log('Deadstorm: hooked');
        this.players[0].setAutofireTimings(39, 39); // 39 + 39 + 6 = 83ms
        this.players[1].setAutofireTimings(39, 39);
    }

    unhook () {
        this.players[0].stopAutofire();
        this.players[0].stopRumblePulsingEffect("GUNFIRE");
        console.log("Deadstorm: unhooked");
    }

    tick () {
        let state = this.#readMemory();
        this.#doP1(state);
        this.#doP2(state);
        this.#lastState = state;
    }

    #doP1 (state) {
        let now = new Date();
        if (state.p1vibration !== this.#lastState.p1vibration && state.p1down) {
            if (!this.#p1FireStart) {
                this.#p1FireStart = now;
                console.log("start af");
                // this.players[0].setAutofireTimings(39, 39);
                this.players[0].stopAutofire();
                this.players[0].startAutofire();
                this.players[0].stopRumblePulsingEffect();
                this.players[0].startRumblePulsingEffect("GUNFIRE", 255, 35, 48);
            }
        }
        if (state.p1vibration === this.#lastState.p1vibration && this.#p1FireStart && (now - this.#p1LastSwitch > 135)) { // (this.#p1Weapon === 1 ? 95 : 95))) {
            console.log('delay STOP AF: ' + (now - this.#p1LastSwitch));
            this.#p1Weapon = 0;
            this.#p1FireStart = null;
            this.#p1LastSwitch = null;
            this.players[0].stopAutofire();
            this.players[0].stopRumblePulsingEffect("GUNFIRE");
        }
        if (state.p1vibration !== this.#lastState.p1vibration && this.#p1FireStart) {
            console.log(now - this.#p1LastSwitch);
            if (now - this.#p1LastSwitch < 55 && now - this.#p1FireStart > 80 && this.#p1Weapon !== 1) {
                // this.players[0].stopAutofire();
                this.players[0].setAutofireTimings(22, 22);
                // this.players[0].startAutofire();
                this.#p1Weapon = 1;
                console.log("1  FAST");
            } else if (now - this.#p1LastSwitch > 50 && (now - this.#p1LastSwitch < 100) && this.#p1Weapon !== 0) {
                // this.players[0].stopAutofire();
                this.players[0].setAutofireTimings(39, 39);
                this.players[0].stopRumblePulsingEffect();
                this.players[0].startRumblePulsingEffect("GUNFIRE", 255, 50, 0);
                // this.players[0].startRumblePulsingEffect("GUNFIRE", 255, 40, 0);
                // this.players[0].startAutofire();
                this.#p1Weapon = 0;
                console.log("0  REGULAR ")
            } else if (now - this.#p1LastSwitch > 200 && this.#p1FireStart !== now && this.#p1Weapon !== 2) {
                this.players[0].setAutofireTimings(0, 9999);
                this.players[0].fireRecoil();
                this.players[0].stopRumblePulsingEffect("GUNFIRE")
                this.players[0].setRumbleEffect("CANNON", 255, 100);
                this.players[0].setRumbleEffect("CANNON2", 150, 200);
                this.players[0].setRumbleEffect("CANNON3", 75, 400);
                this.#p1Weapon = 2;
                console.log("2 CANNON")
            }
        }
        if (!state.p1down && this.#p1FireStart) {
            console.log("stop af: " + state.p1down + ',' + this.#p1FireStart);
            this.#p1Weapon = 0;
            this.#p1FireStart = null;
            this.#p1LastSwitch = null;
            this.players[0].stopAutofire();
            this.players[0].setAutofireTimings(39, 39);
            this.players[0].stopRumblePulsingEffect("GUNFIRE");
        }
        if (state.p1vibration !== this.#lastState.p1vibration && state.p1down) {
            this.#p1LastSwitch = new Date();
        }
    }

    #doP2 (state) {
        let now = new Date();
        if (state.p2vibration !== this.#lastState.p2vibration && state.p2down) {
            if (!this.#p2FireStart) {
                this.#p2FireStart = now;
                console.log("start af");
                // this.players[1].setAutofireTimings(39, 39);
                this.players[1].stopAutofire();
                this.players[1].startAutofire();
                this.players[1].stopRumblePulsingEffect();
                this.players[1].startRumblePulsingEffect("GUNFIRE", 255, 35, 48);
            }
        }
        if (state.p2vibration === this.#lastState.p2vibration && this.#p2FireStart && (now - this.#p2LastSwitch > 135)) { // (this.#p2Weapon === 1 ? 95 : 95))) {
            console.log('delay STOP AF: ' + (now - this.#p2LastSwitch));
            this.#p2Weapon = 0;
            this.#p2FireStart = null;
            this.#p2LastSwitch = null;
            this.players[1].stopAutofire();
            this.players[1].stopRumblePulsingEffect("GUNFIRE");
        }
        if (state.p2vibration !== this.#lastState.p2vibration && this.#p2FireStart) {
            console.log(now - this.#p2LastSwitch);
            if (now - this.#p2LastSwitch < 55 && now - this.#p2FireStart > 80 && this.#p2Weapon !== 1) {
                // this.players[1].stopAutofire();
                this.players[1].setAutofireTimings(22, 22);
                // this.players[1].startAutofire();
                this.#p2Weapon = 1;
                console.log("1  FAST");
            } else if (now - this.#p2LastSwitch > 50 && (now - this.#p2LastSwitch < 100) && this.#p2Weapon !== 0) {
                // this.players[1].stopAutofire();
                this.players[1].setAutofireTimings(39, 39);
                this.players[1].stopRumblePulsingEffect();
                this.players[1].startRumblePulsingEffect("GUNFIRE", 255, 50, 0);
                // this.players[1].startRumblePulsingEffect("GUNFIRE", 255, 40, 0);
                // this.players[1].startAutofire();
                this.#p2Weapon = 0;
                console.log("0  REGULAR ")
            } else if (now - this.#p2LastSwitch > 200 && this.#p2FireStart !== now && this.#p2Weapon !== 2) {
                this.players[1].setAutofireTimings(0, 9999);
                this.players[1].fireRecoil();
                this.players[1].stopRumblePulsingEffect("GUNFIRE")
                this.players[1].setRumbleEffect("CANNON", 255, 100);
                this.players[1].setRumbleEffect("CANNON2", 150, 200);
                this.players[1].setRumbleEffect("CANNON3", 75, 400);
                this.#p2Weapon = 2;
                console.log("2 CANNON")
            }
        }
        if (!state.p2down && this.#p2FireStart) {
            console.log("stop af: " + state.p2down + ',' + this.#p2FireStart);
            this.#p2Weapon = 0;
            this.#p2FireStart = null;
            this.#p2LastSwitch = null;
            this.players[1].stopAutofire();
            this.players[1].setAutofireTimings(39, 39);
            this.players[1].stopRumblePulsingEffect("GUNFIRE");
        }
        if (state.p2vibration !== this.#lastState.p2vibration && state.p2down) {
            this.#p2LastSwitch = new Date();
        }
    }

    #readMemory() {
        let outputs = this.mem.read(this.#outputsAddress, TYPE.UINT32);
        let triggerDown = this.mem.read(0x301348BBC + 0x1, TYPE.UINT32);
        let ret = {
            p1vibration: outputs & 0xC0,
            p1down: triggerDown & 0x80,
            p2vibration: outputs & 0x30,
            p2down: triggerDown & 0x10
        };
        return ret;
    }
}

// todo -- pass in config & md5 info.
exports.generate = (process, players, mem) => { return new DeadstormPlugin(process, players, mem); };