/* Feedback plugin for HOTD4, v0.1.2.
 Created by nugarp/spazzy. Last Updated 10/03/22.
 Big thanks to argonlefou for help with the
 memory addresses. No requirement on loader but
 only tested with TP.

 v0.1.2 - Force set autofire timings in hook,
  and made the grenade affect both players when
  it explodes. Also fixed typo for p2 grenade.
 v0.1.1 - Fixed typo with "died" condition.
 */

const { Plugin } = require('./plugin');
const { TYPE } = require('../core/memoryHelpers');

class Hotd4Plugin extends Plugin {

    // these variables are updated in hook
    #playerStructPtrAddress  = 0x0;
    #outputsPtrAddress = 0x0;
    #gameModePtrAddress = 0x0;

    // temporary lastState to use at initialization
    #lastState = {
        mode1: -1,
        life1: -1,
        ammo1: -1,
        gren1: -1,
        mode2: -1,
        life2: -1,
        ammo2: -1,
        gren2: -1,
        motor1: 0,
        motor2: 0
    };

    constructor(proc, players, mem, cfg) {
        console.log('Hot4d: loaded');
        super(proc, players, mem, cfg);
    }

    hook () {
        console.log('Hot4d: hooked');
        let checkInstructionRevA = 0x08152851;
        let checkInstructionRevC = 0x08153065;
        let revAId = this.mem.readBuffer(checkInstructionRevA, 5);
        let revCId = this.mem.readBuffer(checkInstructionRevC, 5);
        if (revCId[0] === 0xE8 && revCId[1] === 0x42 && revCId[2] === 0x0D && revCId[3] === 0 && revCId[4] === 0) {
            this.#playerStructPtrAddress  = 0x0A6F27F8;
            this.#outputsPtrAddress = 0x0A6F2754;
            this.#gameModePtrAddress = 0x0A6F27A8;
            console.log("Hot4d: Rev. C matched");
        } else if (revAId[0] === 0xE8 && revAId[1] === 0x42 && revAId[2] === 0x0D && revAId[3] === 0 && revAId[4] === 0) {
            this.#playerStructPtrAddress  = 0x0A7117A8;
            this.#outputsPtrAddress = 0x0A711704;
            this.#gameModePtrAddress = 0x0A711758;
            console.log("Hot4d: Rev. A matched");
        } else {
            console.log("Unknown Hot4d executable");
            throw "Unknown Hot4d executable. Addresses are not supported.";
        }
        this.players[0].setAutofireTimings(22, 22); // 22 + 22 + 6 = 50ms
        this.players[1].setAutofireTimings(22, 22);
    }

    unhook () {
        console.log("Hot4d: unhooked");
    }

    tick () {
        let state = this.#readMemory();
        let lastState = this.#lastState;
        if (state.gameMode == 8 && state.valid) {

            // map player properties to the full state object
            let playersStates = [
                { mode: "mode1", life: "life1", ammo: "ammo1", gren: "gren1", motor: "motor1" },
                { mode: "mode2", life: "life2", ammo: "ammo2", gren: "gren2", motor: "motor2" },
            ];

            // check and set conditions for each player
            for(let i = 0; i < playersStates.length; i++) {
                let cp = playersStates[i]; // current player
                let device = this.players[i]; // serial device interface
                let op = playersStates[(i + 1) % 2]; // the other player - used for grenade explosion rumble
                let odevice = this.players[(i + 1) % 2]; // the other device - used for grenade explosion rumble

                if (state[cp.mode] == 11) {
                    if (state[cp.life] == 0 && state[cp.life] < lastState[cp.life]) { // got killed
                        console.log("-1 life -- you died!!");
                        device.setRumbleEffect("DIED", 255, 600);
                    }
                } else if (state[cp.mode] === 3 || state[cp.mode] === 4) {
                    if (state[cp.life] < lastState[cp.life]) {
                        console.log("-1 life");
                        device.setRumbleEffect("HIT", 255, 250);
                    }
                    if (state[cp.ammo] < lastState[cp.ammo]) {
                        console.log("-" + (lastState[cp.ammo] - state[cp.ammo]) + " ammo");
                    } else if (state[cp.ammo] > lastState[cp.ammo]) {
                        console.log("RELOAD");
                        device.setRumbleEffect("RELOAD", 255, 150);
                    }
                    if (state[cp.gren] < lastState[cp.gren]) {
                        setTimeout(() => {
                            device.setRumbleEffect("GRENADE", 255, 600);
                            if (state[op.life] > 0) {
                                odevice.setRumbleEffect("GRENADE", 255, 600);
                            }
                        }, 850);
                        console.log("GRENADE");
                    }
                }
                if (state[cp.life] === 0 && state[cp.life] < lastState[cp.life]) {
                    console.log("possibly another place you died?");
                    device.setRumbleEffect("DIED", 255, 600);
                }
                if (state[cp.motor] !== lastState[cp.motor]) {
                    if (state[cp.motor] === 1) {
                        device.startAutofire();
                        device.setRumbleEffect("AUTOFIRE", 175);
                    } else {
                        device.stopAutofire();
                        device.setRumbleEffect("AUTOFIRE", 0);
                    }
                }
            }
        }
        this.#lastState = state;
    }

    #readMemory() {
        let outputsAddress = this.mem.read(this.#outputsPtrAddress, TYPE.UINT32);
        let motor1 = this.mem.read(outputsAddress, TYPE.BYTE) >> 6 & 0x01;
        let motor2 = this.mem.read(outputsAddress, TYPE.BYTE) >> 3 & 0x01;
        this.GameModePtr = this.mem.read(this.mem.read(this.#gameModePtrAddress, TYPE.UINT32) + 0x2C, TYPE.UINT32);
        let gameMode = this.mem.read(this.GameModePtr + 0x38, TYPE.BYTE);
        let ret = { 
            mode1: this.#lastState.mode1,
            life1: this.#lastState.life1,
            ammo1: this.#lastState.ammo1,
            gren1: this.#lastState.gren1,
            mode2: this.#lastState.mode2,
            life2: this.#lastState.life2,
            ammo2: this.#lastState.ammo2,
            gren2: this.#lastState.gren2,
            valid: false,
            motor1: motor1,
            motor2: motor2,
            gameMode: gameMode
        };

        if (gameMode == 8) {
            let playersPtrBaseAddress = this.mem.read(this.#playerStructPtrAddress, TYPE.UINT32);
            if (playersPtrBaseAddress !== 0) {
                let p1StructAddress  = this.mem.read(playersPtrBaseAddress + 0x34, TYPE.UINT32);
                let p2StructAddress  = this.mem.read(playersPtrBaseAddress + 0x38, TYPE.UINT32);
                let mode1 = this.mem.read(p1StructAddress + 0x38, TYPE.BYTE);
                let life1 = this.mem.read(p1StructAddress + 0x3C, TYPE.BYTE);
                let ammo1 = this.mem.read(p1StructAddress + 0x274, TYPE.BYTE);
                let gren1 = this.mem.read(p1StructAddress + 0x2A0, TYPE.BYTE);
                let mode2 = this.mem.read(p2StructAddress + 0x38, TYPE.BYTE);
                let life2 = this.mem.read(p2StructAddress + 0x3C, TYPE.BYTE);
                let ammo2 = this.mem.read(p2StructAddress + 0x274, TYPE.BYTE);
                let gren2 = this.mem.read(p2StructAddress + 0x2A0, TYPE.BYTE);
    
                ret = {
                    mode1: mode1,
                    life1: life1,
                    gren1: gren1,
                    ammo1: ammo1,
                    motor1: motor1,
                    mode2: mode2,
                    life2: life2,
                    ammo2: ammo2,
                    gren2: gren2,
                    motor2: motor2,
                    valid: true,
                    gameMode: gameMode
                }
            }
        }
        return ret;
    }
}

// todo -- pass in config & md5 info.
exports.generate = (process, players, mem) => { return new Hotd4Plugin(process, players, mem); };