/* Feedback plugin for LA Machineguns, v0.1.1.
 Created by nugarp/spazzy. Last Updated 9/26/22.
 Model3 outputs a static value for when the motor
 is held down...we want autofire! Recoil and rumble!
 Thhis is an example plugin that does just that by
 leveraging model3's outputs. Make sure you run the
 rom with the parameter -outputs=win (or is it output=win)?

 v0.1.1: fixed p2 output.

 Overall, a very simple plugin.
 */
 
const { HookPlugin } = require('./hookplugin')

// HookPlugins can leverage the STATE object in ticks:
// state["Property"] = { value, oldValue, changed }
// state.changed = boolean indicating ANY change since last poll

class LaMachinPlugin extends HookPlugin {
    constructor(info, players, nil) {
        console.log('LaMachin loaded')
        super(info, players, ["LampView2", "LampView3"]) // register relevant outputs
    }

    hook () {
        console.log("LaMachin: hooked");
        this.players[0].setAutofireTimings(40,74); // 40 + 74 + 6 = 120ms
        this.players[1].setAutofireTimings(40,74); // 40 + 74 + 6 = 120ms
    }

    tick () {
        if (!this.state.changed) {
            return;
        }
        let vars = ["LampView2", "LampView3"];
 
        // if LampView "i" has changed, setRumble/start/stopAutofire for player "i".
        for (let i = 0; i < vars.length; i++) {
            if (this.state[vars[i]].changed) {
                if (this.state[vars[i]].value === 1) {
                    this.players[i].startAutofire();
		            this.players[i].setRumbleEffect("GUN-FIRST", 255, 100); // make the first shot a little longer in case it was a single shot
                    this.players[i].startRumblePulsingEffect("GUN", /* strength */ 255, /* on period */ 60, /* off period */ 60);
                } else {
                    // we dont immediately stop rumble for GUN-FIRST because we want at least one shot to rumble.
                    this.players[i].stopAutofire();
                    this.players[i].stopRumblePulsingEffect("GUN");
                }
            }
        }
        super.tick();
    }
    
    unhook () {
        console.log("LaMachin: unhooked");
    }
}

exports.generate = (info, players, nil) => { return new LaMachinPlugin(info, players, nil); };