/* Feedback plugin for Let's Go Island 3D, v0.1.
 Created by nugarp/spazzy. Last Updated 9/26/22.
 This is an example of a file that is using MAMEHooker outputs
 from another program (in this case, TP/Output Blaster) in order
 to transform the motor effects to rumble + recoil. */

const { HookPlugin } = require('./hookplugin')

// HookPlugins can leverage the STATE object in ticks:
// state["Property"] = { value, oldValue, changed }
// state.changed = boolean indicating ANY change since last poll

class LGI3DPlugin extends HookPlugin {
    constructor(info, players, nil) {
        console.log("Let's Go Island 3D - OutputBlaster - Loaded")
        super(info, players, ["1pMotor", "2pMotor", "1pKnock", "2pKnock"]); // register relevant outputs
    }

    hook () {
        console.log("Let's Go Island 3D - Output Blaster - Hooked");
        this.players[0].setAutofireTimings(40,37); // 40 + 37 + 6 = 83ms
        this.players[1].setAutofireTimings(40,37); // 40 + 37 + 6 = 83ms
    }

    #knockId = 0;

    tick () {
        let motors = ["1pMotor", "2pMotor"];
        let knocks = ["1pKnock", "2pKnock"];
        motors.forEach((m, i) => {
            if (this.state[m].changed && this.state[m].value) {
                this.players[i].startAutofire();
                this.players[i].setRumbleEffect("GUN", 175);
            } else if (this.state[m].changed && !this.state[m].value) {
                this.players[i].stopAutofire();
                this.players[i].setRumbleEffect("GUN", 0);
            }
        });
        knocks.forEach((k, i) => {
            if (this.state[k].changed && this.state[k].value) {
                // we can allow for unlimited knocks so we'll suffix it with a random ID
                this.players[i].setRumbleEffect("KNOCK" + this.#knockId, 255, 150);
            }
        });
        super.tick();
    }

    unhook () {
        this.players[0].setAutofireTimings(45,45);
        this.players[1].setAutofireTimings(45,45);
        console.log("Let's Go Island 3D - Output Blaster - Unhooked");
    }
}



exports.generate = (info, players, nil) => { return new LGI3DPlugin(info, players, nil); };