/* Feedback plugin for Point Blank (MAME), v0.1.1.
 Created by nugarp/spazzy. Last Updated 12/7/22.
 v0.1.1 - removed 4:3 / 16:9 setter in connected() and unhook().
 instead, people can override this logic with the overrides folder.
 */

const { Plugin } = require('./plugin');
const { TYPE } = require('../core/memoryHelpers');

class PtBlankPlugin extends Plugin {

    // these variables are updated in hook
    #gameBaseAddr = null;
    #id = 0;

    // temporary lastState to use at initialization
    #lastState =  { players: [ { bullets: -1, life: -1, wtarget: -1, bomb: -1 }, { bullets: -1, life: -1, wtarget: -1, bomb: -1 } ], changed: false, valid: false };

    constructor(proc, players, mem, cfg) {
        console.log('ptblank: loaded');
        super(proc, players, mem, cfg);
    }

    hook () {
        console.log('ptblank: hooked');
        this.players[0].setAutofireTimings(30, 30); // 30 + 30 + 6 = 66ms (just under 4 frames, which is 67ms)
        this.players[1].setAutofireTimings(30, 30);
        this.#id = 0;
    }

    unhook () {
        console.log("ptblank: unhooked");
    }

    tick () {
        let state = this.#readMemory();

        if (state.valid) {
            if (state.changed) {
                // console.log(state);
                for (let i = 0; i < state.players.length; i++) {
                    let device = this.players[i]; // serial device
                    let player = state.players[i]; // player info (ammo, life, bullets)
                    let lastPlayer = this.#lastState.players[i];
    
                //     // health-related
                //     if (player.life === 0 && lastPlayer.life === 1) { // died
                //         device.setRumbleEffect("DIED", 255, 600);
                //     } else if (player.life === lastPlayer.life - 1) { // got hit
                //         device.setRumbleEffect("HIT", 255, 250);
                //     } else if (player.life > lastPlayer.life) { // picked up a health bar or maybe started a new game
                //         device.setRumbleEffect("HEALTH-UP", 255, 100); // same feedback as reload..just a short 'tick'
                //     }
    
                    // shooting-related
                    // shot a bullet...check either ammo ctr goes down or total shot ctr goes up
                    if (player.life > 0 && (player.bullets == lastPlayer.bullets + 1)) {
                        device.fireRecoil(1);
                        device.setRumbleEffect("GUN-RECOIL", 255, 100); // originally strong
                        device.setRumbleEffect("GUN-RECOIL2", 175, 125); // ...then fade it out
                        device.setRumbleEffect("GUN-RECOIL3", 100, 150); // ...most faded
                    }
    
                    // hit a bomb target (or the target belonging to the other player)
                    if (player.life > 0 && (player.wtarget == lastPlayer.wtarget + 1 || player.bomb == lastPlayer.bomb + 1)) {
                        device.setRumbleEffect("BOMB" + this.#id++, 255, 350);
                        device.setRumbleEffect("BOMB" + this.#id++, 200, 400);
                        device.setRumbleEffect("BOMB" + this.#id++, 150, 450);
                        device.setRumbleEffect("BOMB" + this.#id++, 100, 500);
                    }
    
                    // lost a life
                    if (player.life >= 0 && player.life < lastPlayer.life) {
                        device.setRumbleEffect("LOST-LIFE" + this.#id++, 255, 200);
                    }

                    // scored a bonus life
                    if (player.life >= 0 && player.life == lastPlayer.life + 1) {
                        device.setRumbleEffect("GAINED-LIFE" + this.#id++, 255, 200);
                    }
                }
            }
            this.#lastState = state;
        }
    }

    #lastPatternCheck = new Date();

    #readMemory() {
        // this is an expensive operation so we only try it if 1 second has elapsed from the last check
        if (!this.#gameBaseAddr) {
            if ((new Date() - this.#lastPatternCheck) > 1000) {
                console.log("[ptblank] checking for base address");
                this.#lastPatternCheck = new Date();
                // can't find a static address so we find this pattern in memory and then the valuable addresses are based off of finding this seq.
                // let addr = this.mem.findPattern('B0 97 21 00 1C 97 21 00 88 96 00 00 F4 95');
                let addr = this.mem.findPattern('B0 97 21 ? 1C 97 21 ?');
                if (addr) {
                    console.log("[ptblank] base address found @ " + (addr - 0x17FA6).toString(16));
                    this.#gameBaseAddr = addr - 0x17FA6;
                } else {
                    return { valid: false };
                }
            } else {
                return { valid: false };
            }
        }

        let lookups = [
            { bullets: this.#gameBaseAddr + 0x17F76, life: this.#gameBaseAddr + 0x8, wtarget: this.#gameBaseAddr, bomb: this.#gameBaseAddr + 0x4 },
            { bullets: this.#gameBaseAddr + 0x17F74, life: this.#gameBaseAddr + 0xE, wtarget: this.#gameBaseAddr + 0x6, bomb: this.#gameBaseAddr + 0xA }
        ];

        // state = { players: [ { ammo, life, wtarget, bomb }, { ammo, life, wtarget, bomb } ], changed, valid }
        let state = {
            players: [{},{}],
            changed: false,
            valid: true
        };

        // everything can be read as a SHORT so we can pretty nicely update everything with this loop
        lookups.forEach((L, i) => { // L = lookup object (i.e. field => address), i = player index (e.g. 0 or 1)
            Object.keys(L).forEach(f => { // f = field (e.g. "ammo", "life", "bullets")
                state.players[i][f] = this.mem.read(lookups[i][f], TYPE.SHORT);
                // console.log(this.#lastState);
                state.changed = state.changed || (state.players[i][f] !== this.#lastState.players[i][f]);
                state.valid = state.valid && state.players[i][f] >= 0; // ignore if values are negative.
            })
        });

        // state.valid = state.valid && state.players[0].life <= 99 && state.players[1].life <= 99; //invalid values are quite high...1326583323, 1179648
        // state.valid = state.valid && state.players[0].bullets <= 65535 && state.players[1].bullets <= 65535 // doubt someone will shoot > 16M bullets. invalid values are 1.3B
        // state.valid = state.valid && state.players[0].ammo <= 5 && state.players[1].ammo >= 0
        return state;
    }
}

// todo -- pass in config & md5 info.
exports.generate = (process, players, mem) => { return new PtBlankPlugin(process, players, mem); };