/* Feedback plugin for Time Crisis 5, v0.1.
 Created by nugarp/spazzy. Last Updated 9/25/22.
 Recommended that you launch TC5 with TP, or else
 you will not have feedback for the rocket launcher
 in stage 2. */

 const { Plugin } = require('./plugin');
 const { TYPE } = require('../core/memoryHelpers');
 const NanoTimer = require('nanotimer');
 
 class Tc5Plugin extends Plugin {
     // stage 1 area 2 has the helicopter mini-scene: handgun ONLY and ammo = 255 (inf)
     // downhill upside bad guy is the same
     // temporary lastState to use at initialization
     // weapons:
     // 0 = handgun
     // 1 = machine gun (max=180)
     // 2 = shotgun
     // 3 = grenade
     // 4 = mounted machine gun (autofire) (stage 2)
     // 5 = rocket launcher? (autofire) (stage 2)
     // 6 = sniper rifle, ammo is always 255 (stage 4 / area 1), single fire
     #lastState = { life: 6, weapon: 0, ammo: [9, 100, 10, 1], changed: false };
 
     constructor(proc, players, mem, cfg) {
         console.log('Tc5: loaded');
         super(proc, players, mem, cfg);
     }
 
     hook () {
         console.log('Tc5: hooked');
         this.players[0].setAutofireTimings(40,54); // 40 + 54 + 6 = 100ms
         this.players[0].sendAction("M8.0"); // turn off full auto -- needed beacuse we will be doing S1 partially
     }
 
     unhook () {
         this.players[0].stopAutofire();
         this.#lastFireRocketInterval.clearInterval();
         this.#lastFireRocketInterval.clearTimeout();
         this.players[0].sendAction("M8.1"); //auto mode
         console.log("Tc5: unhooked");
     }
 
     // autofire for regular auto-weapons (machine guns)
     #isAutofire = false;
     #lastAutofireBulletTime = new Date();
     // autofire data for the damn rockets, which are delayed @ 300ms
     #lastFireRocket = new Date();
     #lastFireRocketInterval = new NanoTimer();
     // vars to ensure that we are in the game
     #lastCheckPlaying = new Date();
     #lastPlaytime = 0;
     #isPlaying = false;
     // just an incrementing id for debugging purposes
     #id = 0;
     tick () {
         let state = this.#readMemory();
         if (new Date() - this.#lastCheckPlaying > 50) { // check every 50ms that playtime has changed..this will tell us if we are "playing"
             this.#isPlaying = state.playtime !== this.#lastPlaytime;
             this.#lastCheckPlaying = new Date();
             this.#lastPlaytime = state.playtime;
         }
 
         if (!state.valid) { return; }
 
         // detect singular firing
         // this logic can likely be simplified after discovering the bulletsFired data point, but I'm too lazy to work up that rework right now.
         // NOT machine gun + ammo for this weapon is lesser than before + NOT ammo value 255 (used for intermediate cutscenes with shooting..ie slowmo)
         if (state.weapon !== 1 && state.ammo[state.weapon] < this.#lastState.ammo[state.weapon] && this.#lastState.ammo[state.weapon] !== 255) { // shot was fired
             this.players[0].fireRecoil(1);
             switch(state.weapon) {
                 case 2:
                     this.players[0].setRumbleEffect("SHOTGUN", 255, 200);
                     break;
                 case 3:
                     this.players[0].setRumbleEffect("GRENADE", 255, 400);
                     break;
                 case 0:
                 default:
                     this.players[0].setRumbleEffect("HANDGUN", 255, 75);
                     break;
                 
             }
             console.log("SINGLE SHOT   " + this.#id++);
 
         // cases to START autofiring:
         // weapon is 4 (mounted machine gun) OR 5 (rocket launcher) AND triggerDown = 1 AND isAutofire = 0 AND either pedal is down 
         // weapon is 1 (machine gun) AND ammo decreased AND isAutofire = 0
         } else if (
             ((state.weapon === 1 && state.ammo[state.weapon] < this.#lastState.ammo[state.weapon])
                 || (state.weapon === 4 && state.bulletsFired > this.#lastState.bulletsFired)
                 || state.weapon === 5 && state.isValidForRocketLauncherFire) // we dont do bulletsFired since that doesnt increment til the rocket explodes....
             && !this.#isAutofire && this.#isPlaying) {
 
             this.#lastAutofireBulletTime = new Date();
             this.#isAutofire = true;
             if (state.weapon === 1 || state.weapon === 4) { // both machine guns go at the same freq - 100ms/bullet
                 console.log("START AUTOFIRE   " + this.#id++);
                 this.players[0].setAutofireTimings(40,54); // 40 + 54 + 6 = 100ms
                 this.players[0].startAutofire(); // 40 + 54 + 6 = 100 
                 this.players[0].setRumbleEffect("AUTOFIRE-FIRST", 255, 75);
                 this.players[0].startRumblePulsingEffect("AUTOFIRE", 175, 50, 50);
             } else if (state.weapon === 5) { // rocket launcher is slower, 300ms/bullet
                 // since we have custom logic for this rocket launcher (i.e. handling rapid fire start/stops in the middle)
                 // we won't use the regular startAutofire(). We'll instead define our own way of doing it.
                 // 300ms gives players a lot of time to trigger press/release between shots, so...
 
                 this.players[0].setAutofireTimings(60, 234); // 60 + 234 + 6 = 300ms
                 let lfrTemp = new Date();
                 let delay = 0;
 
                 // if not enough time has passed since they last hit the rocket button...we'll create a delay.
                 if (lfrTemp - this.#lastFireRocket < 300) { // havent had enough time for the next bullet yet...
                     delay = lfrTemp - this.#lastFireRocket;
                 }
 
                 // custom logic so we can't just use standard .startAutofire();
                 // this code is very similar to the way the startAutofire is implemented natively anyway.
                 this.#lastFireRocketInterval.setTimeout(() => {
                     this.players[0].setAutofireTimings(60, 234); // 60+234+6 = 300 ms..
                     this.#lastFireRocketInterval.setInterval(() => {
                         this.#lastAutofireBulletTime = new Date();
                         console.log("AF SHOT" + this.#id++);
                         this.players[0].fireRecoil(1);
                         this.#lastFireRocket = new Date();
                         this.players[0].setRumbleEffect("ROCKET-LAUNCHER", 255, 225, 75);
                     }, '', '300m');
                     console.log("AF SHOT" + this.#id++);
                     this.#lastAutofireBulletTime = new Date();
                     this.players[0].fireRecoil(1);
                     this.#lastFireRocket = new Date();
                     this.players[0].setRumbleEffect("ROCKET-LAUNCHER", 255, 225, 75);
                 }, '', delay + 'm');
             }
         }
 
         // update the last time a bullet was fired for weapons 1 & 4...for weapon 5...just pretend we're firing it if the following conditions
         // are met (as set in #readMemory: triggerDown & (leftPedal || rightPedal))
         if (this.#isPlaying && 
             (
             (state.weapon === 1 || state.weapon === 4) && state.bulletsFired > this.#lastState.bulletsFired) // 4 = mounted machinegun (unlimited ammo)
             || (state.weapon === 5 && state.isValidForRocketLauncherFire) // 5 = rocket launcher (unlimited ammo)
         ) {
             this.#lastAutofireBulletTime = new Date();
         }
         
         // if the trigger was newly down and we have infinite ammo, file a recoil anyway
         // handgun & sniper rifle are single shot w/inf ammo scenarios
         if (((state.weapon === 0 && state.ammo[0] === 255) || (state.weapon === 6)) && state.bulletsFired > this.#lastState.bulletsFired) {
             console.log("UNLIMITED AMMO RECOIL   " + this.#id++);
             this.players[0].fireRecoil(1);
         }
 
         // STOP autofiring
         // if AUTOFIRE is ON...and...
         // either weapon changed
         // or ammo ran out (non-mounted machine gun ONLY) OR ammo hasn't been shot recently (also only relevant for weapon 1? or can we forget it now that we have trigger release?)
         // or trigger is released
         // or died
         // or not playing
         if (this.#isAutofire &&
             ((state.weapon !== this.#lastState.weapon)
             || (state.weapon === 1 && state.ammo[1] === 0 || (new Date() - this.#lastAutofireBulletTime > 120)) // ammo only relevant for the (non-mounted) machine gun
             // || ((state.weapon === 4 || state.weapon === 5) && !state.triggerDown) // other autofire guns 
             || !state.triggerDown // maybe we always stop on trigger up anyway...
             || (state.life === 0 && this.#lastState.life > 0) // you died
                 || !this.#isPlaying) // not playing
         ) {
             console.log("STOP AUTOFIRE   " + this.#id++);
             this.#isAutofire = false;
             this.players[0].stopAutofire();
             this.players[0].stopRumblePulsingEffect("AUTOFIRE", 0);
             if (state.weapon === 5) {
                 this.#lastFireRocketInterval.clearInterval();
                 this.#lastFireRocketInterval.clearTimeout();
             }
         }
 
         // damage rumble
         if (state.life < this.#lastState.life) {
             console.log("HIT EVENT");
             this.players[0].setRumbleEffect("HIT", 255, 600);
         }
 
         // weapon change rumble
         if (state.weapon !== this.#lastState.weapon) {
             console.log("WEAPON CHANGE");
             this.players[0].setRumbleEffect("WEAPON-CHANGE", 255, 100);
             if (state.weapon === 0 || state.weapon === 2 || state.weapon === 3 || state.weapon === 6) { 
                 this.players[0].setAutofireTimings(30,0); // allow for fast shooting on these manual guns. 
             }
         }
 
         // reload rumble
         if (state.ammo[state.weapon] > this.#lastState.ammo[state.weapon] && state.ammo[state.weapon] <= 180) {
             // we'll delay it by 150ms for animation time.
             this.players[0].setRumbleEffect("RELOAD", 255, 80);
         }
 
         state.changed = false;
         if (state.valid) {
             this.#lastState = state;
         }
     }
 
     #readMemory() {
         // these ptrs only work when using TP. They are only used for rocket launcher feedback...so if using any other launchers
         // then you will only lose out on that part of the feedback.
         let pedalPtr = this.mem.readChain64(BigInt(this.process.modBaseAddr) + 0x0185BE18n, [0x3Cn, 0x20n, 0x68n, 0x558n]) + 0x2F8n;
         let leftPedal = this.mem.read(pedalPtr, TYPE.BYTE) >> 6;
         let rightPedal = this.mem.read(pedalPtr + 0x1n, TYPE.BYTE) >> 1 & 0x01;
         let playtime = this.mem.read(BigInt(this.process.modBaseAddr) + 0x18616B0n, TYPE.FLOAT);
         // let timerptr = this.mem.readChain64(BigInt(this.process.modBaseAddr) + 0x0185BE20n, [0x60n, 0x130n, 0x158n, 0x40n, 0xC4n]) + 0xB4Cn;
         // let timer = this.mem.read(timerptr, TYPE.FLOAT);
 
         let ptrAmmoLife = this.mem.readChain64(BigInt(this.process.modBaseAddr) + 0x0185BE18n, [ 0x1Cn, 0x08n, 0x70n, 0x1FCn ]);
         let triggerDown = this.mem.read(BigInt(this.process.modBaseAddr) + 0x1860413n, TYPE.BYTE);
         let bulletsFired = this.mem.read(BigInt(this.process.modBaseAddr) + 0x18616A0n, TYPE.UINT32);
         if (ptrAmmoLife !== 0) {
             let life = this.mem.read(ptrAmmoLife + 0x10n, TYPE.BYTE) >> 1;
             let weapon = this.mem.read(ptrAmmoLife + 0x118n, TYPE.BYTE);
             let ammo = this.mem.read(ptrAmmoLife + 0x170n, TYPE.BYTE);
 
             // use this ONLY for rocketlauncher since fired delay is until it explodes.
             let isValidForRocketLauncherFire = weapon === 5 && (leftPedal || rightPedal) && triggerDown;
             
             // main state object
             let now = { life: life, weapon: weapon, ammo: this.#lastState.ammo.map(x => x), valid: life >= 0 && life <= 9 && weapon <= 6 && (ammo <= 180 || ammo === 255), triggerDown: triggerDown, isValidForRocketLauncherFire: isValidForRocketLauncherFire, leftPedal: leftPedal, rightPedal: rightPedal, bulletsFired: bulletsFired, playtime: playtime}; // do a deep copy of the ammo array
             
             // store ammo for current weapon
             now.ammo[weapon] = ammo;
             
             // detect changes
             now.changed = (now.life !== this.#lastState.life || now.weapon !== this.#lastState.weapon || now.ammo[weapon] !== this.#lastState.ammo[weapon] || now.triggerDown !== this.#lastState.triggerDown || now.bulletsFired !== this.#lastState.bulletsFired || now.isValidForRocketLauncherFire !== this.#lastState.isValidForRocketLauncherFire || playtime > this.#lastState.playtime);
             
             // if (now.changed) {
             //     console.log(JSON.stringify(now));
             // }
             return now;
         }
         return { life: 3, weapon: 0, ammo: [9, 100, 10, 1] , changed: false, valid: false, triggerDown: 0, isValidForRocketLauncherFire: false, bulletsFired: 0 , playtime: 0 };
     }
 }
 
 // todo -- pass in config & md5 info.
 exports.generate = (process, players, mem) => { return new Tc5Plugin(process, players, mem); };