/* Feedback plugin for Transformers Human Alliance, v0.1.1.
 Created by nugarp/spazzy. Last Updated 10/04/22.
 This is an example of a file that is using MAMEHooker outputs
 from another program (in this case, DemulShooter) in order
 to transform the GunMotor effects to rumble + recoil.
 
 v0.1.1: Reduced autofiregoing check from 160ms to 100ms so
  the guns are more responsive at when they stop autofire
  (i.e. continuation will reevaluate 60ms sooner)
  */

const { HookPlugin } = require('./hookplugin')

// HookPlugins can leverage the STATE object in ticks:
// state["Property"] = { value, oldValue, changed }
// state.changed = boolean indicating ANY change since last poll

class TransformersHAPlugin extends HookPlugin {
    constructor(info, players, nil) {
        console.log('TransformersHA loaded 2')
        super(info, players, ["P1_GunMotor", "P2_GunMotor", "P1_LmpGun", "P2_LmpGun", "P1_Life", "P2_Life"]) // register relevant outputs
    }

    hook () {
        console.log("TransformersHA: hooked");
        this.players[0].setAutofireTimings(40,37); // 40 + 37 + 6 = 83ms
        this.players[1].setAutofireTimings(40,37); // 40 + 37 + 6 = 83ms
    }

    #isAutofire = [false, false];
    #lastGunMotorWasDown = [new Date(), new Date()];

    tick () {
        // unfortunately, we can't rely on checking state changed to shortcut out of the logic here because
        // we are relying on time since GunMotor = 1 for some logic
        // if (!this.state.changed) {
        //     return;
        // }
        let guns = ["P1_GunMotor", "P2_GunMotor"];
        let lamps = ["P1_LmpGun", "P2_LmpGun"];
        let lives = ["P1_Life", "P2_Life"];
        for (let i = 0; i < guns.length; i++) {
            if (this.state[lives[i]].changed && this.state[lives[i]].value < this.state[lives[i]].oldValue) {
                // this is my lazy way to actually create a "decreasing rumble" effect. It starts @ 255 and decreases to 75/255 by 1100ms.
                this.players[i].setRumbleEffect("HIT", 255, 600);
                this.players[i].setRumbleEffect("HIT2", 230, 675);
                this.players[i].setRumbleEffect("HIT3", 205, 750);
                this.players[i].setRumbleEffect("HIT4", 175, 850);
                this.players[i].setRumbleEffect("HIT5", 125, 950);
                this.players[i].setRumbleEffect("HIT6", 75, 1100);
            }

            // this logic is probably a little roundabout but it works...
            // to trigger ON, lamp must be on AND gun must be on
            // to trigger OFF, if lamp is off, turn off
            // to stay ON, ensure last time gun was on was within 100ms.
            let triggerStart = !this.#isAutofire[i] && this.state[guns[i]].value && this.state[lamps[i]].value;
            let keepAutofireGoing = this.#isAutofire[i] && ((new Date() - this.#lastGunMotorWasDown[i]) < 100); // check for 100ms to allow for some lag if it occurs.
            let stopAutofire = this.#isAutofire[i] && !this.state[lamps[i]].value && !keepAutofireGoing;
            stopAutofire = stopAutofire || (this.#isAutofire && !this.state[guns[i]].value && ((new Date() - this.#lastGunMotorWasDown[i]) > 160));

            if (this.state[guns[i]].value) {
                this.#lastGunMotorWasDown[i] = new Date();
            }

            if (triggerStart) {
                this.players[i].startAutofire();
                this.players[i].setRumbleEffect("GUN-FIRST", 255, 80); // make the first shot a little longer in case it was a single shot
                this.players[i].startRumblePulsingEffect("GUN", /* strength */ 255, /* on period */ 40, /* off period */ 43);
                this.#isAutofire[i] = true;
            } else if (stopAutofire) {
                this.players[i].stopAutofire();
                this.players[i].stopRumblePulsingEffect("GUN");
                this.#isAutofire[i] = false;
            }
        }
        super.tick();
    }

    unhook () {
        console.log("TransformersHA: unhooked");
    }
}



exports.generate = (info, players, nil) => { return new TransformersHAPlugin(info, players, nil); };