/* Feedback plugin for Terminator Salvation, v0.1a.
 Created by nugarp/spazzy. Last Updated 2/27/24.
 Recommended that you launch TS with TP, or else
 you will not have feedback for the mounted gun.
 
 0.1a - fixed issue when p1 died but p2 didn't and
 the firing would be continuous for no apparent reason. */

const { Plugin } = require('./plugin');
const { TYPE } = require('../core/memoryHelpers');

WEAPONS = {
    SHOTGUN: 0,
    DEFAULT: 1,
    MINIGUN: 2,
    MOUNTED: 5,
    MENU: 7
}
WEAPON_TIMINGS = {
    0: [60, 234],
    1: [40, 54],
    2: [23, 23],
    5: [40, 54],
    7: [0, 150]
}
WEAPON_INTERVALS = {
    0: 200,
    1: 100,
    2: 50,
    5: 100,
    7: 200
}

class TsPlugin extends Plugin {
    // weapons:
    // 0 = shotgun // 300ms - 315ms? => 60 + 234 + 6 = 100ms
    // 1 = default // 100ms => 40 + 54 + 6 = 100ms
    // 2 = minigun // 50ms => 23 + 23 + 6 = 50ms
    // 5 = mounted gun // 100ms 
    // 7 = menu gun // no feedback?


    #lastState = { players: [{ life: 0, weapon: 1, ammo: 0, grenades: 0, shots: 0 }, { life: 0, weapon: 1, ammo: 0, grenades: 0, shots: 0 }], changed: false };
    
    constructor(proc, players, mem, cfg) {
        console.log('TS: loaded');
        super(proc, players, mem, cfg);
    }

    hook () {
        console.log('TS: hooked');
        this.players[0].setAutofireTimings(40,54); // 40 + 54 + 6 = 100ms
        this.players[0].sendAction("M8.0"); // turn off full auto -- needed beacuse we will be doing S1 partially
        if (this.players[1]) {
            this.players[1].setAutofireTimings(40,54); // 40 + 54 + 6 = 100ms
            this.players[1].sendAction("M8.0"); // turn off full auto -- needed beacuse we will be doing S1 partially    
        }
    }

    unhook () {
        this.players[0].stopAutofire();
        this.players[0].sendAction("M8.1"); //auto mode
        if (this.players[1]) {
            this.players[1].stopAutofire();
            this.players[1].sendAction("M8.1"); //auto mode    
        }
        console.log("TS: unhooked");
    }

    #isAutofire = { players: [false, false] }
    #lastAutofireBulletTime = { players: [new Date(), new Date()] }

    tick () {
        let state = this.#readMemory();
        // we'll treat individual players as valid instead of the entire state.
        // if (!state.valid) { return; }

        let numPlayers = Math.min(this.players.length, 2)
        for (let playerId = 0; playerId < numPlayers; playerId++) {

            let shotWasGrenade = state.players[playerId].shots > this.#lastState.players[playerId].shots && state.players[playerId].grenades == this.#lastState.players[playerId].grenades - 1 && state.players[playerId].valid

            if (shotWasGrenade) {
                this.players[playerId].setRumbleEffect("GRENADE", 255, 200);
                this.players[playerId].setRumbleEffect("GRENADE-2", 150, 300);
                this.players[playerId].setRumbleEffect("GRENADE-3", 75, 400);
            }

            // cases to start autofiring:
            // shots changed AND isAutofire = 0 AND player is valid
            if (!shotWasGrenade && state.players[playerId].shots > this.#lastState.players[playerId].shots && !this.#isAutofire.players[playerId] && state.players[playerId].valid) {
                console.log("START AUTOFIRE")
                this.#lastAutofireBulletTime.players[playerId] = new Date()
                this.#isAutofire.players[playerId] = true
                let weapon = state.players[playerId].weapon
                let timings = WEAPON_TIMINGS[weapon]
                let interval = WEAPON_INTERVALS[weapon]
                this.players[playerId].setAutofireTimings(timings[0], timings[1])
                this.players[playerId].startAutofire()
                this.players[playerId].setRumbleEffect("AUTOFIRE-FIRST", 255, 75);
                this.#lastAutofireBulletTime.players[playerId] = new Date()
                if (weapon === WEAPONS.SHOTGUN) {
                    // big rumble for the shotgun
                    this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 255, 225, 75);
                    this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-1", 255, 150); // in case it gets cut off early.
                    this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-2", 150, 200); // in case it gets cut off early.
                    this.players[playerId].setRumbleEffect("AUTOFIRE-SHOTGUN-3", 100, 225); // in case it gets cut off early.
                } else if (weapon === WEAPONS.MINIGUN) {
                    this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 125, 30, 20);
                } else {
                    this.players[playerId].startRumblePulsingEffect("AUTOFIRE", 175, 50, interval - 50);
                }
            }

            if (!shotWasGrenade && state.players[playerId].shots > this.#lastState.players[playerId].shots && state.players[playerId].valid) {
                this.#lastAutofireBulletTime.players[playerId] = new Date()
            }        

            // stop autofire if player (1) weapon change, (2) dead, (3) hasn't fired in the interval specified
            if (this.#isAutofire.players[playerId] &&
                (state.players[playerId].weapon !== this.#lastState.players[playerId].weapon || // weapon change
                state.players[playerId].life <= 0 || state.players[playerId].life > 1000 || // dead
                !state.players[playerId].valid || // also effectively dead...
                ((new Date() - this.#lastAutofireBulletTime.players[playerId]) > WEAPON_INTERVALS[state.players[playerId].weapon]))
                ) {
                this.players[playerId].stopAutofire();
                this.players[playerId].stopRumblePulsingEffect("AUTOFIRE", 0);
                this.#isAutofire.players[playerId] = false
                this.#lastAutofireBulletTime.players[playerId] = 0
            }

            // // damage rumble
            if (state.players[playerId].life < this.#lastState.players[playerId].life) { // we don't check valid because they could have just died.
                console.log("HIT EVENT");
                this.players[playerId].setRumbleEffect("HIT", 255, 125);
            }

            // // weapon change rumble
            if (state.players[playerId].weapon !== this.#lastState.players[playerId].weapon && state.players[playerId].valid) {
                let weapon = state.players[playerId].weapon
                console.log("WEAPON CHANGE");
                this.players[playerId].setRumbleEffect("WEAPON-CHANGE", 255, 100);
                switch(weapon) {
                    // 0 = shotgun // 300ms - 315ms? => 60 + 234 + 6 = 100ms
                    // 1 = default // 100ms => 40 + 54 + 6 = 100ms
                    // 2 = minigun // 50ms => 23 + 23 + 6 = 50ms
                    // 5 = mounted gun // 100ms 
                    // 7 = menu gun // no feedback?
                    // TODO use timings variable
                    case 0: // shotgun
                        this.players[playerId].setAutofireTimings(60, 234);
                        break
                    case 1: // default
                    case 5: // mounted gun
                        this.players[playerId].setAutofireTimings(40, 54);
                        break
                    case 2: // minigun
                        this.players[playerId].setAutofireTimings(23, 23);
                        break
                    default:
                        this.players[playerId].setAutofireTimings(40, 54);
                }
            }

            // reload rumble
            if (state.players[playerId].ammo > this.#lastState.players[playerId].ammo && state.players[playerId].weapon === this.#lastState.players[playerId].weapon && state.players[playerId].valid) {
                this.players[playerId].setRumbleEffect("RELOAD", 255, 80);
            }
        }

        if (state.changed) { 
            console.log(JSON.stringify(state))
        }

        this.#lastState = state
    }

    #readMemory() {
        // thanks to Argon's DemulShooter for p1/p2 life, ammo, and weapon # addresses
        let p1LifePtr = 0x088CC240 + 0x58
        let p1WeaponPtr = 0x88A5F80
        let p1AmmoPtr = 0x88DBF68
        let p1ShotsPtr = 0x088AD7C8
        let p1GrenadePtr = 0x088DBF6C
        // let p1BulletsPtr = 0x088CC2AC // not used

        let p2LifePtr = 0x088CC2BC + 0x58
        let p2WeaponPtr = 0x88A5FE0
        let p2AmmoPtr = 0x88DBF7C
        let p2ShotsPtr = 0x088AD7C4
        let p2GrenadePtr = 0x088DBF80
        // let p2BulletPtr = 0x088CC328 // not used

        let p1Life = this.mem.read(p1LifePtr, TYPE.INT32)
        let p1Weapon = this.mem.read(p1WeaponPtr, TYPE.INT32)
        let p1Ammo = this.mem.read(p1AmmoPtr, TYPE.INT32)
        let p1Shots = this.mem.read(p1ShotsPtr, TYPE.INT32)
        let p1Grenades = this.mem.read(p1GrenadePtr, TYPE.INT32)

        let p2Life = this.mem.read(p2LifePtr, TYPE.INT32)
        let p2Weapon = this.mem.read(p2WeaponPtr, TYPE.INT32)
        let p2Ammo = this.mem.read(p2AmmoPtr, TYPE.INT32)
        let p2Shots = this.mem.read(p2ShotsPtr, TYPE.INT32)
        let p2Grenades = this.mem.read(p2GrenadePtr, TYPE.INT32)

        let now = { players: [
            { life: p1Life, weapon: p1Weapon, ammo: p1Ammo, grenades: p1Grenades, shots: p1Shots, valid: p1Life >=0 && p1Life <= 1000 },
            { life: p2Life, weapon: p2Weapon, ammo: p2Ammo, grenades: p2Grenades, shots: p2Shots, valid: p2Life >=0 && p2Life <= 1000 }
        ], changed: false, valid: false };

        now.valid = now.players[0].valid || now.players[1].valid

        for (let i = 0; i < 2; i++) {
        now.changed = now.changed ||
            now.players[i].life !== this.#lastState.players[i].life ||
            now.players[i].weapon !== this.#lastState.players[i].weapon ||
            now.players[i].shots !== this.#lastState.players[i].shots ||
            now.players[i].grenades !== this.#lastState.players[i].grenades

        }

        return now
    }
}

// todo -- pass in config & md5 info.
exports.generate = (process, players, mem) => { return new TsPlugin(process, players, mem); };